# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.functions import Power, Lagrange, DG, Composite, Tree


class Nurbs(Tree):
    def __init__(self, dimRange=1):
        Tree.__init__(self, "Nurbs")
        self.dimRange = dimRange

    def __repr__(self):
        if self.dimRange == 1:
            return "Nurbs"
        else:
            return "Nurbs^" + str(self.dimRange)


duneFunctionsLayouts = {"lexicographic": "Lexicographic", "interleaved": "Interleaved"}


def indexMergingStrategy(blocked, layout):
    return (
        "Dune::Functions::BasisBuilder::"
        + ("Blocked" if blocked else "Flat")
        + duneFunctionsLayouts[layout]
    )


def preBasisTypeName(tree, gridViewTypeName):
    assert isinstance(tree, Tree)
    if isinstance(tree, Lagrange):
        scalarPreBasis = (
            "Dune::Functions::LagrangePreBasis< "
            + gridViewTypeName
            + " , "
            + str(tree.order)
            + " >"
        )
        if tree.dimRange != 1:
            IMS = indexMergingStrategy(False, "interleaved")
            return (
                "Dune::Functions::PowerPreBasis< "
                + IMS
                + " , "
                + scalarPreBasis
                + " , "
                + str(tree.dimRange)
                + " >"
            )
        else:
            return scalarPreBasis
    elif isinstance(tree, Nurbs):
        scalarPreBasis = "Dune::Functions::NurbsPreBasis< " + gridViewTypeName + " >"
        if tree.dimRange != 1:
            IMS = indexMergingStrategy(False, "interleaved")
            return (
                "Dune::Functions::PowerPreBasis< "
                + IMS
                + " , "
                + scalarPreBasis
                + " , "
                + str(tree.dimRange)
                + " >"
            )
        else:
            return scalarPreBasis
    elif isinstance(tree, DG):
        raise Exception(repr(tree) + " not supported by dune-functions.")
    elif isinstance(tree, Composite):
        IMS = indexMergingStrategy(tree.blocked, tree.layout)
        ChildPreBases = " , ".join(
            preBasisTypeName(c, gridViewTypeName) for c in tree.children
        )
        return (
            "Dune::Functions::CompositePreBasis< " + IMS + " , " + ChildPreBases + " >"
        )
    elif isinstance(tree, Power):
        IMS = indexMergingStrategy(tree.blocked, tree.layout)
        ChildPreBasis = preBasisTypeName(tree.children[0], gridViewTypeName)
        return (
            "Dune::Functions::PowerPreBasis< "
            + IMS
            + " , "
            + ChildPreBasis
            + " , "
            + str(tree.exponent)
            + " >"
        )
    else:
        raise Exception("Unknown type of tree: " + repr(tree))


from dune.functions import load


def defaultGlobalBasis(gridView, tree):
    headers = [
        "powerbasis",
        "compositebasis",
        "lagrangebasis",
        "subspacebasis",
        "defaultglobalbasis",
    ]

    includes = []
    includes += ["dune/functions/functionspacebases/" + h + ".hh" for h in headers]
    includes += list(gridView.cppIncludes)
    includes += ["dune/iga/nurbsbasis.hh"]
    typeName = (
        "Dune::Functions::DefaultGlobalBasis< "
        + preBasisTypeName(tree, gridView.cppTypeName)
        + " >"
    )

    return load(includes, typeName).GlobalBasis(gridView)



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later
from ._iga import *
from ._nurbspatchdata import *
from ._igagrids import *
from .basis import *
from ._boundarypatch import *
from ._nurbsAlgorithms import *

registry = dict()
registry["grid"] = grid_registry
#
registry["globalBasis"] = {"IGA": defaultGlobalBasis}



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later
from dune.common.hashit import hashIt
from dune.common import FieldVector
from dune.generator.generator import SimpleGenerator


def ControlPoint(coords, weight=1):
    fv = FieldVector(coords)
    generator = SimpleGenerator("ControlPoint", "Dune::Python")

    element_type = f"Dune::IGA::ControlPoint<{fv.cppTypeName}>"

    includes = []
    includes += ["dune/python/iga/nurbspatchdata.hh"]
    moduleName = "ControlPoint_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )

    return module.ControlPoint(fv, weight)


def ControlPointNet(controlPoints):
    generator = SimpleGenerator("MultiDimensionNet", "Dune::Python")

    try:
        controlPointType= controlPoints[0][0][0].cppTypeName
        netDim=3
    except:
        try:
            controlPointType= controlPoints[0][0].cppTypeName
            netDim=2
        except:
            try:
                controlPointType= controlPoints[0].cppTypeName
                netDim=1
            except:
                raise Exception("Controlpoint type not deducable from list")
    element_type = f"Dune::IGA::MultiDimensionNet<{netDim},{controlPointType}>"

    includes = []
    includes += ["dune/python/iga/nurbspatchdata.hh"]
    moduleName = "NurbsPatchData_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )

    return module.MultiDimensionNet(controlPoints)


def NurbsPatchData(knotSpans, controlPointNet, degree):
    generator = SimpleGenerator("NurbsPatchData", "Dune::Python")

    worldDim = len(controlPointNet.directGet( 0).coords)
    dim = controlPointNet.netDim
    element_type = f"Dune::IGA::NURBSPatchData<{dim},{worldDim},double>"

    includes = []
    includes += ["dune/python/iga/nurbspatchdata.hh"]
    moduleName = "NurbsPatchData_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )

    if isinstance(knotSpans,tuple):
        if isinstance(knotSpans[0],float) or  isinstance(knotSpans[0], int):
            knotSpans= list([list(knotSpans),])
    if isinstance(degree,int):
        degree=(degree,)
    return module.NurbsPatchData(knotSpans, controlPointNet, degree)


def NurbsPatchDataDefault(dim, worldDim):
    generator = SimpleGenerator("NurbsPatchData", "Dune::Python")

    element_type = f"Dune::IGA::NURBSPatchData<{dim},{worldDim},double>"
    includes = []
    includes += ["dune/python/iga/nurbspatchdata.hh"]
    moduleName = "NurbsPatchData_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )

    return module.NurbsPatchData()



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.common.hashit import hashIt
from dune.generator.generator import SimpleGenerator


def boundaryPatch(gridView, booleanVector):
    generator = SimpleGenerator("BoundaryPatch", "Dune::IGA::Python")
    element_type = f"BoundaryPatch<{gridView.cppTypeName}>"

    includes = []

    includes += ["dune/python/iga/boundarypatch.hh"]
    includes += gridView._includes
    moduleName = "boundaryPatch_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    return module.BoundaryPatch(gridView, booleanVector)



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.generator.algorithm import run
from dune.generator import path
from io import StringIO
from dune.generator.importclass import load
from dune.iga import NurbsPatchDataDefault

runCircularArc = """
#include <dune/iga/nurbsalgorithms.hh>
auto run( double radius = 1.0,  double startAngle = 0.0, double endAngle = 360.0,
                       const Dune::FieldVector<double, 3>& origin = {0, 0, 0},
                       const Dune::FieldVector<double, 3>& X      = {1, 0, 0},
                       const Dune::FieldVector<double, 3>& Y      = {0, 1, 0})
{
  return Dune::IGA::makeCircularArc(radius,startAngle,endAngle,origin,X,Y);
}
"""


def makeCircularArc(
    radius, startAngle=None, endAngle=None, origin=None, X=None, Y=None
):
    NurbsPatchDataDefault(dim=1, worldDim=3)
    if startAngle is None:
        return run("run", StringIO(runCircularArc), radius)
    elif endAngle is None:
        return run("run", StringIO(runCircularArc), radius, startAngle)
    elif origin is None:
        return run("run", StringIO(runCircularArc), radius, startAngle, endAngle)
    elif X is None:
        return run(
            "run", StringIO(runCircularArc), radius, startAngle, endAngle, origin
        )
    elif Y is None:
        return run(
            "run", StringIO(runCircularArc), radius, startAngle, endAngle, origin, X
        )
    else:
        return run(
            "run", StringIO(runCircularArc), radius, startAngle, endAngle, origin, X, Y
        )


runSurfaceOfRevolution = """
#include <dune/iga/nurbsalgorithms.hh>
auto run(const Dune::IGA::NURBSPatchData<1, 3, double>& generatrix,
                               const Dune::FieldVector<double, 3>& point,
                               const Dune::FieldVector<double, 3>& revolutionaxisI,
                               const double& revolutionAngle = 360.0)
{
  return Dune::IGA::makeSurfaceOfRevolution(generatrix,point,revolutionaxisI,revolutionAngle);
}
"""


def makeSurfaceOfRevolution(generatrix, point, revolutionaxisI, revolutionAngle=None):
    if generatrix.patchDim != 1 and generatrix.dimworld != 3:
        raise ValueError(
            f"The passed nurbs patchdata has to be a curve in 3D. You passed a {generatrix.patchDim}-dimensional object in {generatrix.dimworld}D space"
        )
    from dune.common import FieldVector

    NurbsPatchDataDefault(dim=2, worldDim=3)
    pointFV = FieldVector(point)
    revolutionaxisIFV = FieldVector(revolutionaxisI)
    print(generatrix)
    if revolutionAngle is None:
        return run(
            "run",
            StringIO(runSurfaceOfRevolution),
            generatrix,
            pointFV,
            revolutionaxisIFV,
        )
    else:
        return run(
            "run",
            StringIO(runSurfaceOfRevolution),
            generatrix,
            pointFV,
            revolutionaxisIFV,
            revolutionAngle,
        )



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later
import sys
import setpath
import math

setpath.set_path()

import os

# enable DUNE_SAVE_BUILD to test various output options
os.environ["DUNE_LOG_LEVEL"] = "debug"
os.environ["DUNE_SAVE_BUILD"] = "console"
os.environ["ALUGRID_VERBOSITY_LEVEL"] = "0"

from dune.iga import (
    IGAGrid,
    ControlPointNet,
    ControlPoint,
    NurbsPatchData,
    makeCircularArc,
    makeSurfaceOfRevolution,
)
from dune.grid import structuredGrid

from dune.iga import reader as readeriga
from dune.iga.basis import defaultGlobalBasis, Power, Lagrange, Nurbs
from dune.common import FieldVector
from dune.grid import gridFunction

if __name__ == "__main__":
    # one dimensional test
    cp = ControlPoint((0, 0, 0), 1)
    cp2 = ControlPoint((0, 0, 3), 1)
    netC = (((cp, cp2)))
    net = ControlPointNet(netC)
    nurbsPatchData = NurbsPatchData(((0, 0, 1, 1)), net, (1))
    gridView = IGAGrid(nurbsPatchData)
    reader = (readeriga.json, "../../iga/test/auxiliaryFiles/element_trim.ibra")
    refinements = 5
    gridView = IGAGrid(reader, dimgrid=2, dimworld=2)
    gridView.hierarchicalGrid.globalRefine(refinements)
    fv = FieldVector(
        10 * [0.1]
    )  # this is here due to https://gitlab.dune-project.org/core/dune-common/-/issues/340

    globalBasis = defaultGlobalBasis(gridView, Power(Nurbs(), 2))

    vtkWriter = gridView.trimmedVtkWriter()

    gf1 = gridView.function(
        lambda e, x: math.sin(
            math.pi * (e.geometry.toGlobal(x)[0] + e.geometry.toGlobal(x)[1])
        )
    )

    gridView.writeVTK("test", pointdata=[gf1])

    @gridFunction(gridView)
    def g(x):
        return [math.sin(2 * math.pi * x[0] * x[1]), x[0] * x[1]] * 5

    vtkWriter.addPointData(g, name="g")
    vtkWriter.write(name="TestGrid")

    cp = ControlPoint((1, 2), 3)
    cp2 = cp + cp

    assert cp.coords[0] == 1
    assert cp.coords[1] == 2
    assert cp.weight == 3

    netC = ((cp, cp2), (cp, cp))
    net = ControlPointNet(netC)
    print(net.get((0, 1)))
    print(net.get((1, 0)))
    net.set((1, 0), ControlPoint((1, 4), 3))
    assert net.get((1, 0)).coords[0] == 1
    assert net.get((1, 0)).coords[1] == 4

    nurbsPatchData = NurbsPatchData(((0.0, 0, 1, 1), (0, 0, 1, 1)), net, (1, 1))

    gridView = IGAGrid(nurbsPatchData)
    assert gridView.size(0) == 1
    assert gridView.size(1) == 4
    assert gridView.size(2) == 4
    globalBasis = defaultGlobalBasis(gridView, Power(Nurbs(), 2))
    assert len(net) * 2 == len(globalBasis)
    newnurbsPatchData = nurbsPatchData.degreeElevate(0, 3)
    # degree elevate changes nothing
    gridView = IGAGrid(newnurbsPatchData)
    assert gridView.size(0) == 1
    assert gridView.size(1) == 4
    assert gridView.size(2) == 4
    newnurbsPatchData = nurbsPatchData.knotRefinement((0.5,), 0)
    newnurbsPatchData = newnurbsPatchData.knotRefinement((0.5,), 1)
    gridView = IGAGrid(newnurbsPatchData)
    assert gridView.size(0) == 4
    assert gridView.size(2) == 9

    r = 1.0
    R = 2.0
    circleData = makeCircularArc(r)
    nurbsPatchDataArc = makeSurfaceOfRevolution(circleData, (R, 0, 0), (0, 1, 0), 360.0)
    gridView = IGAGrid(nurbsPatchDataArc)
    vtkWriter = gridView.vtkWriter(subsampling=4)
    vtkWriter.write(name="Torus")

    try:
        nurbsPatchDataArc = makeSurfaceOfRevolution(
            newnurbsPatchData, (R, 0, 0), (0, 1, 0), 360.0
        )
        raise Exception(f"makeSurfaceOfRevolution should have raised an Value error")
    except ValueError:
        pass

    if True:
        reader = (readeriga.json, "../../iga/test/auxiliaryFiles/element.ibra")
        gridView = IGAGrid(reader, dimgrid=2, dimworld=2)

        assert gridView.size(0) == 1
        assert gridView.size(1) == 4
        assert gridView.size(2) == 4
        gridView.hierarchicalGrid.globalRefine(1)
        gridView = gridView.hierarchicalGrid.leafView
        assert gridView.size(0) == 4
        assert gridView.size(2) == 9
        gridView.hierarchicalGrid.globalRefine(1)
        gridView = gridView.hierarchicalGrid.leafView
        assert gridView.size(0) == 16
        assert gridView.size(2) == 25

        assert gridView.dimGrid == 2
        assert gridView.dimWorld == 2

        # read and refine
        inputParameter = dict(
            file_path="../../iga/test/auxiliaryFiles/element.ibra",
            reader=readeriga.json,
            elevate_degree=(1, 1),
        )
        gridView2 = IGAGrid(inputParameter, dimgrid=2, dimworld=2)
        # degree elevation shouldn't change anything
        assert gridView2.size(0) == 1
        assert gridView2.size(1) == 4
        assert gridView2.size(2) == 4

        inputParameter = dict(
            file_path="../../iga/test/auxiliaryFiles/element.ibra",
            reader=readeriga.json,
            pre_knot_refine=(1, 1),
        )
        gridView3 = IGAGrid(inputParameter, dimgrid=2, dimworld=2)
        # degree elevation shouldn't change anything
        assert gridView3.size(0) == 4
        assert gridView3.size(2) == 9

        inputParameter = dict(
            file_path="../../iga/test/auxiliaryFiles/element.ibra",
            reader=readeriga.json,
            post_knot_refine=(1, 1),
        )
        gridView4 = IGAGrid(inputParameter, dimgrid=2, dimworld=2)
        # degree elevation shouldn't change anything
        assert gridView4.size(0) == 4
        assert gridView4.size(2) == 9



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later
# The top part of this code is taken from https://gitlab.dune-project.org/staging/dune-functions/-/blob/master/dune/python/test/poisson.py
# Thus, see also their license https://gitlab.dune-project.org/staging/dune-functions/-/blob/master/COPYING
import setpath

setpath.set_path()
import numpy as np
import math
from scipy.sparse import lil_matrix
import scipy as sp
import dune.geometry
import dune.grid
from dune.iga import reader as readeriga
from dune.iga.basis import defaultGlobalBasis, Power, Lagrange, Nurbs
from dune.iga import IGAGrid, boundaryPatch
from dune.grid import gridFunction
from dune.common import FieldVector
import os

os.environ["ALUGRID_VERBOSITY_LEVEL"] = "0"

# f(x) = 1
f = lambda x: 1

# g(x) = 0
g = lambda x: 0


# boundary of the unit square
def isDirichlet(x):
    return x[0] <= 0.0 or x[0] >= 1.0 or x[1] >= 1.0 or x[1] <= 0.0


# TODO: This assembler loop is very inefficient in terms of run time and should be improved using Python vectorization.
# See discussion at https://gitlab.dune-project.org/staging/dune-functions/-/merge_requests/295 for hints and code pointers.
def localAssembler(element, localBasis):
    n = len(localBasis)

    localA = np.zeros((n, n))
    localB = np.zeros(n)

    # choose a high enough quadrature order
    quadOrder = 3

    # create a quadrature rule and integrate
    quadRule = dune.geometry.quadratureRule(element.type, quadOrder)
    for pt in quadRule:
        pos = pt.position

        # evaluate the local basis functions (this is an array!)
        phi = localBasis.evaluateFunction(pos)

        # evaluate the local basis reference Jacobians (array of arrays)
        phiRefGradient = localBasis.evaluateJacobian(pos)

        # get the transformation matrix
        jacobianInverseTransposed = element.geometry.jacobianInverseTransposed(pos)

        # |det(jacobianInverseTransposed)|
        integrationElement = element.geometry.integrationElement(pos)

        # transform the reference Jacobians to global geometry
        phiGradient = [
            np.dot(jacobianInverseTransposed, np.array(g)[0]) for g in phiRefGradient
        ]

        posGlobal = element.geometry.toGlobal(pos)

        for i in range(n):
            for j in range(n):
                # compute grad(phi_i) * grad(phi_j)
                localA[i, j] += (
                    pt.weight
                    * integrationElement
                    * np.dot(phiGradient[i], phiGradient[j])
                )

            localB[i] += pt.weight * integrationElement * phi[i] * f(posGlobal)

    return localA, localB


def globalAssembler(basis):
    N = len(basis)

    A = lil_matrix((N, N))
    b = np.zeros(N)

    # mark all Dirichlet DOFs (first sqrt(N) are fixed)
    dichichletDOFs = np.zeros(N, dtype=bool)
    dichichletDOFs[: math.floor(math.sqrt(N))] = True

    # interpolate the boundary values
    gCoeffs = np.zeros(N)
    # basis.interpolate(gCoeffs,g)

    # extract grid and localView
    localView = basis.localView()
    grid = basis.gridView

    for element in grid.elements:
        # assign the localView to the current element
        localView.bind(element)

        # set of all shape functions with support in this element
        localBasis = localView.tree().finiteElement.localBasis

        localN = len(localBasis)

        localA, localb = localAssembler(element, localBasis)

        # copy the local entries into the global matrix using the
        # index mapping given by the localView
        for i in range(localN):
            gi = localView.index(i)[0]

            if dichichletDOFs[gi]:
                A[gi, gi] = 1.0
                b[gi] = gCoeffs[gi]
            else:
                b[gi] += localb[i]
                for j in range(localN):
                    gj = localView.index(j)[0]
                    A[gi, gj] += localA[i, j]

    return A, b


############################### START ########################################

# create a trimmed grid from file
reader = (readeriga.json, "../../iga/test/auxiliaryFiles/element_trim_xb.ibra")

gridView = IGAGrid(reader, dimgrid=2, dimworld=2)
gridView.hierarchicalGrid.globalRefine(2)

neumannVertices = np.zeros(gridView.size(2), dtype=bool)

boundaryPatch = boundaryPatch(gridView, neumannVertices)

basis = defaultGlobalBasis(gridView, Nurbs())

# compute A and b
A, b = globalAssembler(basis)

# convert A to Compressed Sparse Column format
Acsc = A.tocsc()

# solve!
x = sp.sparse.linalg.spsolve(Acsc, b)

vtkWriter = gridView.trimmedVtkWriter()
discreteFunction = basis.asFunction(x)

vtkWriter.addPointData(discreteFunction, name="g")
vtkWriter.write(name="Poisson")

xTest = [
    0.0,
    0.0,
    0.0,
    0.0,
    0.60278885,
    0.70725846,
    2.04974976,
    2.00540431,
    1.86744155,
    1.45648652,
    1.31236311,
    2.04982725,
    2.01589274,
    1.91911023,
    1.81531395,
    1.79925452,
    2.04866857,
    2.01931155,
    1.9486634,
    1.88186179,
    1.85137355,
]

assert np.linalg.norm(x - xTest) < 1e-7



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

try:
    from dune.packagemetadata import metaData
except ImportError:
    from packagemetadata import metaData
from skbuild import setup

# When building a new package, update the version numbers below and run:
# access  docker container with mounted repo to /tmp/dune-iga
# build _iga
# cd /tmp/dune-iga
# /dune/dune-common/build-cmake/run-in-dune-env pip install twine scikit-build
# git config --global --add safe.directory /tmp/dune-iga
# /dune/dune-common/build-cmake/run-in-dune-env python setup.py sdist
# /dune/dune-common/build-cmake/run-in-dune-env python -m twine upload dist/* --verbose

duneigaVersion = "0.1.6"
duneVersion = "2.9.0"

metadata = metaData(duneVersion)[1]
metadata["version"] = duneigaVersion

setup(**metadata)



# SPDX-FileCopyrightText: 2023 The dune-iga developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later


def IGAGrid(constructor, dimgrid=None, dimworld=None):
    """
    Create an IGAGrid instance.

    Parameters:
    -----------

        constructor  means of constructing the grid, i.e. a grid reader or a
                     dictionary holding macro grid information
        dimgrid      dimension of grid, i.e. 2 or 3
        dimworld     dimension of world, i.e. 2 or 3 and >= dimension

    Returns:
    --------

    An IGAGrid instance
    """

    if hasattr(constructor, "patchDim"):
        dimgrid = constructor.patchDim

    if hasattr(constructor, "dimworld"):
        dimworld = constructor.dimworld

    if dimgrid == None and dimworld == None:
        raise Exception(
            "If you don't pass the patch data you have to pass dimgrid and dimworld"
        )

    typeName = (
        "Dune::IGA::NURBSGrid< " + str(dimgrid) + ", " + str(dimworld) + ",double>"
    )

    includes = ["dune/python/iga/grid.hh"]
    from dune.generator.generator import SimpleGenerator
    from dune.common.hashit import hashIt

    generator = SimpleGenerator("HierarchicalGrid", "Dune::Python::IGA")
    moduleName = "NURBSGrid_" + hashIt(typeName)
    kwargs = dict()
    kwargs["dynamicAttr"] = True
    kwargs["holder"] = "std::shared_ptr"

    gridModule = generator.load(
        includes=includes, typeName=typeName, moduleName=moduleName, **kwargs
    )

    if type(constructor) is dict:
        readGrid = gridModule.reader(constructor)
    elif type(constructor) is tuple:
        if (len(constructor)) == 2:
            readGrid = gridModule.reader(
                dict(reader=constructor[0], file_path=constructor[1])
            )
        else:
            raise Exception(
                "Only tuple of size two are allowed   (readeriga.json, filename)"
            )
    else:
        readGrid = gridModule.HierarchicalGrid(constructor)
    gridView = gridModule.LeafGrid(readGrid)
    return gridView


grid_registry = {
    "IGA": IGAGrid,
}
